"""
Static variables that we do not expect to change
"""
# Tick data column properties
TIMESTAMP_UNIT = "ms"

# Time bars column names
OPEN = "open"
LOW = "low"
HIGH = "high"
CLOSE = "close"
BASE_ASSET_VOLUME = "base_asset_volume"
BASE_ASSET_BUY_VOLUME = "base_asset_buy_volume"
BASE_ASSET_SELL_VOLUME = "base_asset_sell_volume"
QUOTE_ASSET_VOLUME = "quote_asset_volume"

FIRST_TIMESTAMP = "first_timestamp"
TIMESTAMP = "time"  # will be in ms

# Column names for trading
SIZE = "size"
SIDE = "side"
NUMBER_EXPIRATION_BARS_COLUMN = "num_expiration_bars"
EVENT_END_TIME = "event_end_time"
DAILY_VOL = "daily_vol"
STOP_LOSS = "stop_loss"
PROFIT_TAKING = "profit_taking"
RETURN = "return"
LABEL = "label"
T_VALUE = "t_value"
